// Copyright (C) 2018-2023 Intel Corporation
// SPDX-License-Identifier: Apache-2.0
//

#include "gmock/gmock.h"
#include "openvino/op/constant.hpp"
#include "openvino/op/parameter.hpp"
#include "openvino/op/split.hpp"
#include "split_shape_inference.hpp"
#include "utils.hpp"

using namespace ov;
using namespace ov::intel_cpu;
using namespace testing;

using SplitTestParams = std::tuple<ShapeVector,  // Input shapes
                                   int64_t,      // Split axis
                                   size_t,       // Number of splits
                                   StaticShape   // Expected output(s) shape
                                   >;

class SplitStaticShapeInferenceTest : public OpStaticShapeInferenceTest<op::v1::Split>,
                                      public WithParamInterface<SplitTestParams> {
protected:
    void SetUp() override {
        std::tie(input_shapes, axis, num_of_splits, exp_shape) = GetParam();

        output_shapes = ShapeVector();
        arg = std::make_shared<op::v0::Parameter>(element::f32, input_shapes.front().get_shape());
    }

    int64_t axis;
    size_t num_of_splits;
    std::shared_ptr<op::v0::Parameter> arg;
};

INSTANTIATE_TEST_SUITE_P(1d_shapes,
                         SplitStaticShapeInferenceTest,
                         Values(make_tuple(ShapeVector{{0}, {}}, 0, 1, StaticShape({0})),
                                make_tuple(ShapeVector{{1}, {}}, 0, 1, StaticShape({1})),
                                make_tuple(ShapeVector{{2}, {}}, -1, 1, StaticShape({2})),
                                make_tuple(ShapeVector{{2}, {}}, 0, 2, StaticShape({1})),
                                make_tuple(ShapeVector{{4}, {}}, -1, 2, StaticShape({2})),
                                make_tuple(ShapeVector{{9}, {}}, 0, 3, StaticShape({3})),
                                make_tuple(ShapeVector{{15}, {}}, -1, 5, StaticShape({3}))),
                         PrintToStringParamName());

INSTANTIATE_TEST_SUITE_P(multi_dim_shapes,
                         SplitStaticShapeInferenceTest,
                         Values(make_tuple(ShapeVector{{6, 12, 21}, {}}, 0, 6, StaticShape({1, 12, 21})),
                                make_tuple(ShapeVector{{6, 12, 21}, {}}, -1, 3, StaticShape({6, 12, 7})),
                                make_tuple(ShapeVector{{6, 12, 21}, {}}, -2, 2, StaticShape({6, 6, 21})),
                                make_tuple(ShapeVector{{6, 12, 21}, {}}, 2, 7, StaticShape({6, 12, 3})),
                                make_tuple(ShapeVector{{6, 12, 1, 14}, {}}, -1, 7, StaticShape({6, 12, 1, 2}))),
                         PrintToStringParamName());

TEST_P(SplitStaticShapeInferenceTest, shape_inference_empty_const_map) {
    const auto axis_node = std::make_shared<op::v0::Constant>(element::i64, Shape{}, axis);
    op = make_op(arg, axis_node, num_of_splits);

    shape_inference(op.get(), input_shapes, output_shapes);

    EXPECT_EQ(output_shapes.size(), num_of_splits);
    EXPECT_THAT(output_shapes, Each(exp_shape));
}

TEST_P(SplitStaticShapeInferenceTest, shape_inference_with_const_map) {
    const auto axis_node = std::make_shared<op::v0::Parameter>(element::i64, Shape{});
    op = make_op(arg, axis_node, num_of_splits);

    const auto axis_const = std::make_shared<op::v0::Constant>(element::i64, ov::Shape{}, axis);
    const auto axis_tensor = std::make_shared<ngraph::runtime::HostTensor>(axis_const);
    const std::map<size_t, std::shared_ptr<ngraph::runtime::HostTensor>>& constant_data = {{1, axis_tensor}};

    shape_inference(op.get(), input_shapes, output_shapes, constant_data);

    ASSERT_EQ(output_shapes.front(), exp_shape);
}
