// Copyright (C) 2018-2023 Intel Corporation
// SPDX-License-Identifier: Apache-2.0
//

#include "test_utils.h"

#include "intel_gpu/runtime/format.hpp"

using namespace cldnn;

TEST(format, to_string) {
    typedef std::underlying_type<cldnn::format::type>::type format_underlying_type;
    for (format_underlying_type i = 0; i < static_cast<format_underlying_type>(cldnn::format::format_num); i++) {
        cldnn::format fmt = static_cast<cldnn::format::type>(i);
        ASSERT_NO_THROW(fmt.to_string()) << "Can't convert to string format " << i;
    }
}

TEST(format, traits) {
    typedef std::underlying_type<cldnn::format::type>::type format_underlying_type;
    for (format_underlying_type i = 0; i < static_cast<format_underlying_type>(cldnn::format::format_num); i++) {
        cldnn::format fmt = static_cast<cldnn::format::type>(i);
        ASSERT_NO_THROW(cldnn::format::traits(fmt)) << "Can't get traits for format " << i;
    }
}

struct format_adjust_test_params {
    cldnn::format in_format;
    size_t new_rank;
    cldnn::format expected_format;
};

class format_adjust_test : public testing::TestWithParam<format_adjust_test_params> {
public:
    static std::string PrintToString(testing::TestParamInfo<format_adjust_test_params> param_info) {
        auto in_fmt = param_info.param.in_format.to_string();
        auto new_rank = std::to_string(param_info.param.new_rank);
        auto expected_fmt = param_info.param.expected_format.to_string();
        std::string res = "in_fmt=" + in_fmt + "_new_rank=" + new_rank + "_expected_fmt=" + expected_fmt;
        return res;
    }
 };

TEST_P(format_adjust_test, shape_infer) {
    auto p = GetParam();

    if (p.expected_format == cldnn::format::any) {
        cldnn::format fmt = cldnn::format::any;
        ASSERT_ANY_THROW(fmt = format::adjust_to_rank(p.in_format, p.new_rank)) << fmt.to_string();
    } else {
        ASSERT_EQ(format::adjust_to_rank(p.in_format, p.new_rank), p.expected_format);
    }
}

INSTANTIATE_TEST_SUITE_P(smoke, format_adjust_test,
    testing::ValuesIn(std::vector<format_adjust_test_params>{
        {cldnn::format::bfyx,           3, cldnn::format::bfyx},
        {cldnn::format::bfyx,           4, cldnn::format::bfyx},
        {cldnn::format::bfyx,           5, cldnn::format::bfzyx},
        {cldnn::format::bfyx,           6, cldnn::format::bfwzyx},
        {cldnn::format::bfzyx,          4, cldnn::format::bfyx},
        {cldnn::format::bfzyx,          6, cldnn::format::bfwzyx},
        {cldnn::format::bfwzyx,         3, cldnn::format::bfyx},
        {cldnn::format::b_fs_yx_fsv16,  5, cldnn::format::b_fs_zyx_fsv16},
        {cldnn::format::b_fs_zyx_fsv16, 4, cldnn::format::b_fs_yx_fsv16},
        {cldnn::format::fs_b_yx_fsv32,  5, cldnn::format::any},
        {cldnn::format::nv12,           5, cldnn::format::any},
        {cldnn::format::oiyx,           5, cldnn::format::any},
    }),
    format_adjust_test::PrintToString);

struct axes_test_format_params {
    cldnn::format in_format;
    // First : block_idx, Second : block_size
    std::vector<std::pair<size_t, int>> inner_block;
    std::vector<std::pair<size_t, int>> per_axis_block;
};

class axes_test_format : public testing::TestWithParam<axes_test_format_params> {
public:
    static std::string PrintToString(testing::TestParamInfo<axes_test_format_params> param_info) {
        auto input_fmt = param_info.param.in_format.to_string();
        auto blocks = param_info.param.inner_block;
        auto per_axis_blocks = param_info.param.per_axis_block;
        std::string res = "in_fmt = " + input_fmt + " : ";
        for (auto block : blocks) {
            res += " { " + std::to_string(block.first) + ", " + std::to_string(block.second) + "}";
        }

        res += " > ";
        for (auto block : per_axis_blocks) {
            res += " { " + std::to_string(block.first) + ", " + std::to_string(block.second) + "}";
        }

        return res;
    }
 };

 TEST_P(axes_test_format, simple_test) {
    auto param = GetParam();

    auto per_axis_blocks = format::per_axis_block_size(param.in_format);
    ASSERT_EQ(per_axis_blocks.size(), param.per_axis_block.size());
    for (size_t idx = 0; idx < per_axis_blocks.size(); idx++) {
        ASSERT_EQ(per_axis_blocks.at(idx).first, param.per_axis_block.at(idx).first);
        ASSERT_EQ(per_axis_blocks.at(idx).second, param.per_axis_block.at(idx).second);
    }

    auto blocks = format::block_sizes(param.in_format);
    ASSERT_EQ(blocks.size(), param.inner_block.size());
    for (size_t idx = 0; idx < blocks.size(); idx++) {
        ASSERT_EQ(blocks.at(idx).first, param.inner_block.at(idx).first);
        ASSERT_EQ(blocks.at(idx).second, param.inner_block.at(idx).second);
    }
}

INSTANTIATE_TEST_SUITE_P(smoke, axes_test_format,
    testing::ValuesIn(std::vector<axes_test_format_params>{
        {format::os_is_yx_isa8_osv8_isv4,       {{1, 8}, {0, 8}, {1, 4}}, {{1, 32}, {0, 8}}},
        {format::os_is_yx_isa8_osv16_isv4,      {{1, 8}, {0, 16}, {1, 4}}, {{1, 32}, {0, 16}}},
        {format::os_is_yx_osa4_isa8_osv8_isv2,  {{0, 4}, {1, 8}, {0, 8}, {1, 2}}, {{0, 32}, {1, 16}}},
        {format::os_is_yx_osa4_isa8_osv8_isv4,  {{0, 4}, {1, 8}, {0, 8}, {1, 4}}, {{0, 32}, {1, 32}}},
        {format::os_is_zyx_osa4_isa8_osv8_isv2, {{0, 4}, {1, 8}, {0, 8}, {1, 2}}, {{0, 32}, {1, 16}}},
        {format::os_is_zyx_osa4_isa8_osv8_isv4, {{0, 4}, {1, 8}, {0, 8}, {1, 4}}, {{0, 32}, {1, 32}}},
        {format::os_is_yx_osa2_isa8_osv16_isv2, {{0, 2}, {1, 8}, {0, 16}, {1, 2}}, {{0, 32}, {1, 16}}},
        {format::os_is_yx_osa2_isa8_osv16_isv4, {{0, 2}, {1, 8}, {0, 16}, {1, 4}}, {{0, 32}, {1, 32}}},
        {format::os_is_yx_osa2_isa8_osv8_isv2,  {{0, 2}, {1, 8}, {0, 8}, {1, 2}}, {{0, 16}, {1, 16}}},
        {format::os_is_zyx_osa2_isa8_osv8_isv2, {{0, 2}, {1, 8}, {0, 8}, {1, 2}}, {{0, 16}, {1, 16}}},
        {format::os_is_zyx_isa8_osv8_isv4,      {{1, 8}, {0, 8}, {1, 4}}, {{1, 32}, {0, 8}}},
        {format::os_is_zyx_isa8_osv16_isv4,     {{1, 8}, {0, 16}, {1, 4}}, {{1, 32}, {0, 16}}},
        {format::is_os_yx_osa4_isa8_osv8_isv4,  {{0, 4}, {1, 8}, {0, 8}, {1, 4}}, {{0, 32}, {1, 32}}},
        {format::is_os_yx_isa2_osa8_isv8_osv2,  {{1, 2}, {0, 8}, {1, 8}, {0, 2}}, {{1, 16}, {0, 16}}},
        {format::is_os_yx_isa4_osa8_isv8_osv4,  {{1, 4}, {0, 8}, {1, 8}, {0, 4}}, {{1, 32}, {0, 32}}},
        {format::os_is_yx_osv8_isv4,            {{0, 8}, {1, 4}}, {{0, 8}, {1, 4}}},
        {format::os_is_zyx_osv8_isv4,           {{0, 8}, {1, 4}}, {{0, 8}, {1, 4}}},
        {format::os_is_yx_osv8_isv2,            {{0, 8}, {1, 2}}, {{0, 8}, {1, 2}}},
        {format::os_is_zyx_osv8_isv2,           {{0, 8}, {1, 2}}, {{0, 8}, {1, 2}}},
        {format::g_os_is_yx_osa2_isa8_osv8_isv2,  {{0, 2}, {1, 8}, {0, 8}, {1, 2}}, {{0, 16}, {1, 16}}},
        {format::g_os_is_yx_osa4_isa8_osv8_isv4,  {{0, 4}, {1, 8}, {0, 8}, {1, 4}}, {{0, 32}, {1, 32}}},
        {format::g_os_is_zyx_osa4_isa8_osv8_isv4, {{0, 4}, {1, 8}, {0, 8}, {1, 4}}, {{0, 32}, {1, 32}}},
        {format::g_os_is_yx_osa4_isa8_osv8_isv2,  {{0, 4}, {1, 8}, {0, 8}, {1, 2}}, {{0, 32}, {1, 16}}},
        {format::g_os_is_zyx_osa4_isa8_osv8_isv2, {{0, 4}, {1, 8}, {0, 8}, {1, 2}}, {{0, 32}, {1, 16}}},
        {format::g_os_is_yx_osa2_isa8_osv16_isv4, {{0, 2}, {1, 8}, {0, 16}, {1, 4}}, {{0, 32}, {1, 32}}},
        {format::g_os_is_yx_osa2_isa8_osv16_isv2, {{0, 2}, {1, 8}, {0, 16}, {1, 2}}, {{0, 32}, {1, 16}}},
        {format::gs_oiyx_gsv32,                   {{8, 32}}, {{8, 32}}},
        {format::gs_oizyx_gsv32,                  {{8, 32}}, {{8, 32}}},
        {format::gs_oiyx_gsv16,                   {{8, 16}}, {{8, 16}}},
        {format::gs_oizyx_gsv16,                  {{8, 16}}, {{8, 16}}},
    }),
    axes_test_format::PrintToString);
